/**
 * \file    : sdc_driver.c
 * \date    : 30 August 2012
 * \author  : Cesar Fuguet
 *
 * This file defines the driver of a SD Card device using an SPI controller
 */

#include <sdc_driver.h>
#include <utils.h>

#define SYSCLK_FREQ           50000000
#define SDCARD_RESET_ITER_MAX 4

static struct sdcard_dev sdcard;
static struct spi_dev*   spi;

/**
 * \return  void
 *
 * \brief   Enable SD Card select signal
 */
static void _sdc_enable()
{
    spi_ss_assert(sdcard.spi, sdcard.slave_id);
}

/**
 * \return  void
 *
 * \brief   Disable SD Card select signal
 */
static void _sdc_disable()
{
    spi_ss_deassert(sdcard.spi, sdcard.slave_id);
}

/**
 * \param   tick_count: SD Card clock ticks number
 *
 * \return  void
 *
 * \brief   Enable SD Card clock
 *          The tick count is byte measured (1 tick, 8 clock)
 */
static void _sdc_gen_tick(unsigned int tick_count)
{
    register int i = 0;
    while(i++ < tick_count) spi_put_tx(sdcard.spi, 0xFF, 0);
}

/**
 * \param   lba : Position where the block device access
 *                pointer must be move
 *
 * \return  void
 *
 * \brief   Change block device access pointer position
 *  
 * The block device access pointer is relocated in terms of blocks
 */
void _sdc_lseek(unsigned int lba)
{
    sdcard.access_pointer = sdcard.block_length * lba;
}

/**
 * \return  char from the SD card
 *
 * \brief   Get a byte from the SD Card
 */
static unsigned char _sdc_receive_char()
{
    _sdc_gen_tick(1);

    return spi_get_rx(sdcard.spi, 0);
}

/**
 * \return  sdcard response
 *
 * \brief   Wait for a valid response after the send of a command
 *          This function can return if one of the next two conditions are true:
 *           1. Bit valid received
 *           2. Timeout (not valid bit received after SDCARD_COMMAND_TIMEOUT
 *              wait ticks)
 */
static unsigned char _sdc_wait_response()
{
    unsigned char sdcard_rsp;
    register int  iter;

    iter       = 0;
    sdcard_rsp = _sdc_receive_char();
    while (
            (iter < SDCARD_COMMAND_TIMEOUT) &&
            !SDCARD_CHECK_R1_VALID(sdcard_rsp)
          )
    {
        sdcard_rsp = _sdc_receive_char();
        iter++;
    }

    return sdcard_rsp;
}

/**
 * \return  void
 *
 * \brief   Wait data block start marker
 */
static void _sdc_wait_data_block()
{
	while (_sdc_receive_char() != 0xFE);
}

/**
 * \param   index   : SD card CMD index
 * \param   app     : Type of command, 0 for normal command or 1 for
 *                    application specific
 * \param   args    : SD card CMD arguments
 *
 * \return  response first byte
 *
 * \brief   Send command to the SD card
 */
static int _sdc_send_command ( int      index,
                               int      app  ,
                               void *   args ,
                               unsigned crc7 )
{
    unsigned char sdcard_rsp;
    unsigned char * _args;

    _sdc_gen_tick(5);  

    if (app == SDCARD_ACMD)
    {
        spi_put_tx(sdcard.spi, 0x40 | 55         , 0 );/* CMD and START bit */
        spi_put_tx(sdcard.spi, 0x00              , 0 );/* Argument[0]       */
        spi_put_tx(sdcard.spi, 0x00              , 0 );/* Argument[1]       */
        spi_put_tx(sdcard.spi, 0x00              , 0 );/* Argument[2]       */
        spi_put_tx(sdcard.spi, 0x00              , 0 );/* Argument[3]       */
        spi_put_tx(sdcard.spi, 0x01 | (crc7 << 1), 0 );/* END bit           */

        sdcard_rsp = _sdc_wait_response();
        if (SDCARD_CHECK_R1_ERROR(sdcard_rsp))
        {
            return sdcard_rsp;        
        }
    }

    _args = (unsigned char *) args;

    _sdc_gen_tick(1);  

    spi_put_tx(sdcard.spi, 0x40 | index      , 0 );
    spi_put_tx(sdcard.spi, _args[0]          , 0 );
    spi_put_tx(sdcard.spi, _args[1]          , 0 );
    spi_put_tx(sdcard.spi, _args[2]          , 0 );
    spi_put_tx(sdcard.spi, _args[3]          , 0 );
    spi_put_tx(sdcard.spi, 0x01 | (crc7 << 1), 0 );

    return _sdc_wait_response();
}

static int _sdc_open( unsigned int channel )
{
	unsigned char args[4];
	unsigned char sdcard_rsp;
	unsigned int  iter, ersp;

	sdcard.spi      = spi;
	sdcard.slave_id = channel;

	/* 
	* Supply SD card ramp up time (min 74 cycles)
	*/
	_sdc_gen_tick(10);

	/* 
	* Assert slave select signal
	* Send CMD0 (Reset Command)
	* Deassert slave select signal
	*/
	_sdc_enable();

	args[0] = 0;
	args[1] = 0;
	args[2] = 0;
	args[3] = 0;
	sdcard_rsp = _sdc_send_command(0, SDCARD_CMD, args, 0x4A);
	if ( sdcard_rsp != 0x01 )
	{

		_puts("card CMD0 failed ");
		return sdcard_rsp;
	}

	_sdc_disable();
	/*
	 * send CMD8. If card is pre-v2, It will reply with illegal command.
	 * Otherwise we announce sdhc support.
	 */
	_sdc_enable();
	args[0] = 0;
	args[1] = 0;
	args[2] = 0x01;
	args[3] = 0x01;
	sdcard_rsp = _sdc_send_command(8, SDCARD_CMD, args, 0x63);
	if (!SDCARD_CHECK_R1_VALID(sdcard_rsp)) {
		_puts("card CMD8 failed ");
		return sdcard_rsp;
	}
	if (!SDCARD_CHECK_R1_ERROR(sdcard_rsp)) {
		/* no error, command accepted. get whole reply */
		ersp = _sdc_receive_char();
		ersp = (ersp << 8) | _sdc_receive_char();
		ersp = (ersp << 8) | _sdc_receive_char();
		ersp = (ersp << 8) | _sdc_receive_char();
		if ((ersp & 0xffff) != 0x0101) {
			/* voltage mismatch */
			_puts("card CMD8 mismatch: ");
			_putx(ersp);
			return sdcard_rsp;
		}
		_puts("v2 or later ");
		sdcard.sdhc = 1;
	} else if ((sdcard_rsp & SDCARD_R1_ILLEGAL_CMD) == 0) {
		/* other error */
		_puts("card CMD8 error ");
		return sdcard_rsp;
	} else {
		sdcard.sdhc = 0;
	}
	_sdc_disable();
	/* send CMD41, enabling the card */
	_sdc_enable();
	args[0] = sdcard.sdhc ? 0x40: 0;
	args[1] = 0;
	args[2] = 0;
	args[3] = 0;

	iter = 0;
	while( iter++ < SDCARD_COMMAND_TIMEOUT )
	{
		sdcard_rsp = _sdc_send_command(41, SDCARD_ACMD, args, 0x00);
		if( sdcard_rsp == 0x01 )
		{
			continue;
		}

		break;
	}

	_sdc_disable();
	if (sdcard_rsp) {
		_puts("SD ACMD41 failed ");
		return sdcard_rsp;
	}
	if (sdcard.sdhc != 0) {
		/* get the card capacity to see if it's really HC */
		_sdc_enable();
		args[0] = sdcard.sdhc ? 0x40: 0;
		args[1] = 0;
		args[2] = 0;
		args[3] = 0;
        sdcard_rsp = _sdc_send_command(58, SDCARD_CMD, args, 0x00);
		if (sdcard_rsp) {
			_puts("SD CMD58 failed ");
			return sdcard_rsp;
		}
		ersp = _sdc_receive_char();
		ersp = (ersp << 8) | _sdc_receive_char();
		ersp = (ersp << 8) | _sdc_receive_char();
		ersp = (ersp << 8) | _sdc_receive_char();
		if (ersp & 0x40000000) {
			_puts("SDHC ");
		} else {
			sdcard.sdhc = 0;
		}
		_sdc_disable();
	}
	_puts("card detected ");
	return 0;
}

static unsigned int _sdc_set_block_size(unsigned int len)
{
    unsigned char args[4];
    unsigned char sdcard_rsp;
    register int i;

    /*
     * For now, supported block size is 512 bytes
     */
    if (len != 512) return 1;

    /*
     * When using high capacity SDCARD, the block_length is not a number of bytes
     * but a number of blocks (transfer unit)
     */
    if (sdcard.sdhc)
    {
        sdcard.block_length = len / 512;
        return 0;
    }

    for (i = 0; i < 4; i++)
    {
        args[i] = (len >> (32 - (i+1)*8)) & 0xFF;
    }

    _sdc_enable();

    sdcard_rsp = _sdc_send_command(16, SDCARD_CMD, args, 0x00);
    if ( SDCARD_CHECK_R1_ERROR(sdcard_rsp) )
    {
        _sdc_disable();
        return sdcard_rsp;
    }

    _sdc_disable();

    sdcard.block_length = len;

	return 0;
}

unsigned int _sdc_init( unsigned int channel )
{
    spi = (struct spi_dev*) &seg_ioc_base;

    /**
     * Initializing the SPI controller
     */
    _spi_init (
        spi           ,
        200000        , /**< SPI_clk: 200 Khz */
        SYSCLK_FREQ   , /**< Sys_clk          */
        8             , /**< Charlen: 8       */
        SPI_TX_NEGEDGE,
        SPI_RX_POSEDGE
    );

    /**
     * Initializing the SD Card
     */
    unsigned int iter = 0;
    unsigned char sdcard_rsp;
    unsigned int i;

    while(1)
    {
        _puts("Trying to initialize SD card... ");

        sdcard_rsp = _sdc_open( channel );
        if (sdcard_rsp == 0)
        {
            _puts("OK\n");
            break;
        }

        _puts("KO\n");

        for (i = 0; i < 1000; i++);

        if (++iter >= SDCARD_RESET_ITER_MAX)
        {
            _puts("\nERROR: During SD card reset to IDLE state\n"
                 "/ card response = ");
            _putx(sdcard_rsp);
            _puts("\n");
            _exit();
        }
    }

    /**
     * Set the block length of the SD Card
     */
    sdcard_rsp = _sdc_set_block_size(512);
    if (sdcard_rsp)
    {
        _puts("ERROR: During SD card blocklen initialization\n");
        _exit();
    }

    /**
     * Incrementing SDCARD clock frequency for normal function
     */
    _spi_init (
        spi         ,
        10000000    , /**< SPI_clk 10 Mhz */
        SYSCLK_FREQ , /**< Sys_clk        */
        -1          , /**< Charlen: 8     */
        -1          ,
        -1
    );

    _puts("Finish block device initialization\n\r");

    return 0;
}


///////////////////////////////////////////////////////////////////////////////
//     _sdc_read()
// Transfer data from the block device to a memory buffer. 
// - mode     : BOOT / KERNEL / USER
// - lba      : first block index on the block device
// - buffer   : base address of the memory buffer (must be word aligned)
// - count    : number of blocks to be transfered.
// Returns 0 if success, > 0 if error.
///////////////////////////////////////////////////////////////////////////////
unsigned int _sdc_read( unsigned int mode,
                        unsigned int lba,
                        paddr_t      buffer,
                        unsigned int count )
{
    unsigned char args[4];
    unsigned char sdcard_rsp;
    unsigned int i;
    unsigned int curr = lba;
    unsigned int last = lba + count;

    for ( ; curr < last ; curr++ )
    {
        _sdc_lseek(curr);

        for (i = 0; i < 4; i++)
        {
            args[i] = (sdcard.access_pointer >> (32 - (i+1)*8)) & 0xFF;
        }

        _sdc_enable();

        sdcard_rsp = _sdc_send_command(17, SDCARD_CMD, args, 0x00);
        if ( SDCARD_CHECK_R1_ERROR(sdcard_rsp) )
        {
            _sdc_disable();
            return sdcard_rsp;
        }

        _sdc_wait_data_block();

        if (spi_get_data(sdcard.spi, buffer, 512 ))
        {
            _sdc_disable();
            return 1;
        }

        /*
         * Get the CRC16 (comes at the end of the data block)
         */
        _sdc_receive_char(); // first byte
        _sdc_receive_char(); // second byte

        _sdc_disable();

        buffer += 512;
    }

    return 0;
}

unsigned int _sdc_write( unsigned int mode,
                         unsigned int lba,
                         paddr_t      buffer,
                         unsigned int count )
{
	return 0;
}

unsigned int _sdc_get_status( unsigned int channel ,
                              unsigned int* status )
{
    return BLOCK_DEVICE_IDLE;
}

unsigned int _sdc_get_block_size()
{
    if (sdcard.sdhc) return sdcard.block_length*512;
    else             return sdcard.block_length;
}



/*
 * vim: tabstop=4 : shiftwidth=4 : expandtab : softtabstop=4
 */
